from typing import Tuple, Union

from sql_smith.partial.parameter import NullParameter, BoolParameter
from sql_smith.interfaces import StatementInterface


class Parameter(StatementInterface):
    def __init__(self, value: Union[str, int, float]):
        self._params = (value, )

    @classmethod
    def create(cls, value) -> StatementInterface:
        if value is None:
            return NullParameter()
        if isinstance(value, bool):
            return BoolParameter(value)

        return Parameter(value)

    def sql(self, engine: 'EngineInterface') -> str:
        return engine.export_parameter(engine.get_parameter_placeholder())

    def params(self, engine: 'EngineInterface') -> Tuple:
        return self._params



class CanUnionMixin:
    def union(self, right: 'StatementInterface') -> 'UnionQuery':
        from sql_smith.query.union_query import UnionQuery
        return UnionQuery(self._engine, self, right)

    def union_all(self, right: 'StatementInterface') -> 'UnionQuery':
        return self.union(right).all()



from sql_smith.capability import HasFromMixin, HasWhereMixin, HasLimitMixin
from sql_smith.functions import express
from sql_smith.query import AbstractQuery


class DeleteQuery(HasFromMixin, HasWhereMixin, HasLimitMixin, AbstractQuery):
    """Implements the DELETE query."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__(engine)
        self._from = ()
        self._limit = None
        self._where = None

    def as_expression(self) -> 'ExpressionInterface':
        query = self.start_expression()
        query = self._apply_from(query)
        query = self._apply_where(query)
        query = self._apply_limit(query)

        return query

    def start_expression(self) -> 'ExpressionInterface':
        return express('DELETE')



from .can_union import CanUnionMixin
from .has_from import HasFromMixin
from .has_limit import HasLimitMixin
from .has_offset import HasOffsetMixin
from .has_order_by import HasOrderByMixin
from .has_returning import HasReturningMixin
from .has_where import HasWhereMixin

__all__ = [
    'CanUnionMixin',
    'HasFromMixin',
    'HasLimitMixin',
    'HasOffsetMixin',
    'HasOrderByMixin',
    'HasReturningMixin',
    'HasWhereMixin'
]



from sql_smith.capability import HasReturningMixin
from sql_smith.query import UpdateQuery as BaseUpdateQuery


class UpdateQuery(HasReturningMixin, BaseUpdateQuery):
    """Postgres UpdateQuery that supports RETURNING."""
    def as_expression(self) -> 'ExpressionInterface':
        query = super().as_expression()
        query = self._apply_returning(query)
        return query



from sql_smith.engine import CommonEngine
from sql_smith.query.postgres import InsertQuery, UpdateQuery


class PostgresEngine(CommonEngine):
    """A custom engine for Postgres SQL dialect."""
    def make_insert(self) -> 'InsertQuery':
        """Creates a custom Postgres INSERT query.

        The Postgres insert query supports RETURNING.
        """
        return InsertQuery(self)

    def make_update(self) -> 'UpdateQuery':
        """Creates a custom Postgres INSERT query.

        The Postgres update query supports RETURNING.
        """
        return UpdateQuery(self)

    def get_parameter_placeholder(self):
        return '%s'



from abc import abstractmethod

from .statement_interface import StatementInterface


class CriteriaInterface(StatementInterface):
    @abstractmethod
    def and_(self, right: 'CriteriaInterface') -> 'CriteriaInterface':
        raise NotImplementedError('and_ must be implemented')

    @abstractmethod
    def or_(self, right: 'CriteriaInterface') -> 'CriteriaInterface':
        raise NotImplementedError('or_ must be implemented')



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class Identifier(StatementInterface):
    def __init__(self, name: str):
        self._name = name

    def sql(self, engine: 'EngineInterface') -> str:
        return engine.escape_identifier(self._name)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return ()



from sql_smith.capability \
    import HasFromMixin, HasWhereMixin, HasLimitMixin, HasOffsetMixin, HasOrderByMixin, CanUnionMixin
from sql_smith.functions import identify_all, express, listing, identify
from .abstract_query import AbstractQuery


class SelectQuery(
    CanUnionMixin, HasFromMixin, HasOrderByMixin, HasWhereMixin, HasLimitMixin, HasOffsetMixin, AbstractQuery
):
    """Implements a SELECT query."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__(engine)
        self._from = ()
        self._limit = None
        self._offset = None
        self._order_by = []
        self._where = None

        self._distinct = False
        self._columns = []
        self._joins = []
        self._group_by = []
        self._having = None

    def distinct(self, state: bool = True) -> 'SelectQuery':
        """Add or remove DISTINCT."""
        self._distinct = state
        return self

    def columns(self, *columns) -> 'SelectQuery':
        """Set the columns to select."""
        self._columns = identify_all(*columns)
        return self

    def add_columns(self, *columns):
        """Add columns to the selection."""
        self._columns = (*self._columns, *identify_all(*columns))
        return self

    def join(self, table: str, criteria: 'CriteriaInterface', join_type: str = '') -> 'SelectQuery':
        """Add a join."""
        sql = '{} JOIN {{}} ON {{}}'.format(join_type.upper()).strip()
        self._joins.append(express(sql, identify(table), criteria))
        return self

    def inner_join(self, table: str, criteria: 'CriteriaInterface') -> 'SelectQuery':
        """Add an INNER join."""
        return self.join(table, criteria, 'INNER')

    def left_join(self, table: str, criteria: 'CriteriaInterface') -> 'SelectQuery':
        """Add a LEFT join."""
        return self.join(table, criteria, 'LEFT')

    def right_join(self, table: str, criteria: 'CriteriaInterface') -> 'SelectQuery':
        """Add a RIGHT join."""
        return self.join(table, criteria, 'RIGHT')

    def full_join(self, table: str, criteria: 'CriteriaInterface') -> 'SelectQuery':
        """Add a FULL join."""
        return self.join(table, criteria, 'FULL')

    def group_by(self, *columns):
        """Add a GROUP BY clause."""
        self._group_by = identify_all(*columns)
        return self

    def having(self, criteria: 'CriteriaInterface'):
        """Add an HAVING clause."""
        self._having = criteria
        return self

    def as_expression(self) -> 'ExpressionInterface':
        query = self.start_expression()
        query = self.__apply_distinct(query)
        query = self.__apply_columns(query)
        query = self._apply_from(query)
        query = self.__apply_joins(query)
        query = self._apply_where(query)
        query = self.__apply_group_by(query)
        query = self.__apply_having(query)
        query = self._apply_order_by(query)
        query = self._apply_limit(query)
        query = self._apply_offset(query)
        return query

    def start_expression(self) -> 'ExpressionInterface':
        return express('SELECT')

    def __apply_columns(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        if len(self._columns) > 0:
            return query.append('{}', listing(self._columns))
        return query.append('*')

    def __apply_distinct(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        if self._distinct:
            return query.append('DISTINCT')
        return query

    def __apply_joins(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('{}', listing(self._joins, ' ')) if len(self._joins) > 0 else query

    def __apply_group_by(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('GROUP BY {}', listing(self._group_by, ' ')) if len(self._group_by) > 0 else query

    def __apply_having(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('HAVING {}', self._having) if self._having else query



from abc import abstractmethod

from sql_smith.interfaces import QueryInterface, EngineInterface
from sql_smith.query.query import Query


class AbstractQuery(QueryInterface):
    """Base class for queries."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__()
        self._engine = engine

    @abstractmethod
    def as_expression(self) -> 'ExpressionInterface':
        raise NotImplementedError('Must override as_expression')

    @abstractmethod
    def start_expression(self) -> 'ExpressionInterface':
        raise NotImplementedError('Must override start_expression')

    def compile(self) -> 'Query':
        """Compiles the query and returns the Query object"""
        query = self.as_expression()
        return Query(
            query.sql(self._engine),
            query.params(self._engine)
        )

    def sql(self, engine: 'EngineInterface') -> str:
        return self.as_expression().sql(engine)

    def params(self, engine: 'EngineInterface') -> str:
        return self.as_expression().params(engine)



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class QualifiedIdentifier(StatementInterface):
    def __init__(self, *args: Tuple['StatementInterface']):
        self._identifiers = args

    def sql(self, engine: 'EngineInterface') -> str:
        return engine.flatten_sql('.', *self._identifiers)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return engine.flatten_params(*self._identifiers)



from abc import abstractmethod

from .statement_interface import StatementInterface


class ExpressionInterface(StatementInterface):
    @abstractmethod
    def append(self, pattern: str, *replacements: 'StatementInterface') -> 'ExpressionInterface':
        raise NotImplementedError('append must be implemented')



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class LikeContains(StatementInterface):
    def __init__(self, value: str):
        self._value = value

    def sql(self, engine: 'EngineInterface') -> str:
        return '?'

    def params(self, engine: 'EngineInterface') -> Tuple:
        value = engine.escape_like(self._value)
        return '%' + value + '%',



from abc import abstractmethod

from .statement_interface import StatementInterface


class QueryInterface(StatementInterface):
    @abstractmethod
    def as_expression(self) -> 'ExpressionInterface':
        raise NotImplementedError('as_expression must be implemented')

    @abstractmethod
    def compile(self) -> 'Query':
        raise NotImplementedError('compile must be implemented')



from sql_smith.functions import literal


class HasOffsetMixin:
    def offset(self, new_offset: int = None):
        self._offset = new_offset
        return self

    def _apply_offset(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('OFFSET {}', literal(self._offset)) if self._offset else query



from sql_smith.capability import HasReturningMixin
from sql_smith.query import InsertQuery as BaseInsertQuery


class InsertQuery(HasReturningMixin, BaseInsertQuery):
    """Postgres InsertQuery that supports RETURNING."""
    def as_expression(self) -> 'ExpressionInterface':
        query = super().as_expression()
        query = self._apply_returning(query)
        return query



class HasWhereMixin:
    def where(self, criteria: 'CriteriaInterface'):
        self._where = criteria
        return self

    def and_where(self, criteria: 'CriteriaInterface'):
        if self._where is None:
            return self.where(criteria)

        self._where = self._where.and_(criteria)
        return self

    def or_where(self, criteria: 'CriteriaInterface'):
        if self._where is None:
            return self.where(criteria)

        self._where = self._where.or_(criteria)
        return self

    def _apply_where(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('WHERE {}', self._where) if self._where else query



from typing import Tuple

from sql_smith.interfaces import ExpressionInterface, EngineInterface


class Expression(ExpressionInterface):
    def __init__(self, pattern: str, *args: 'StatementInterface'):
        self._pattern = pattern
        self._replacements = args

    def append(self, pattern: str, *args: 'StatementInterface') -> 'ExpressionInterface':
        merged = [*self._replacements, *args]
        return Expression(
            '{} {}'.format(self._pattern, pattern),
            *merged
        )

    def sql(self, engine: 'EngineInterface') -> str:
        return self._pattern.format(*map(engine.extract_sql(), self._replacements))

    def params(self, engine: 'EngineInterface') -> Tuple:
        return engine.flatten_params(*self._replacements)



from sql_smith.functions import literal


class HasLimitMixin:
    def limit(self, new_limit: int = None):
        self._limit = new_limit
        return self

    def _apply_limit(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('LIMIT {}', literal(self._limit)) if self._limit else query



from .basic_engine import BasicEngine
from .common_engine import CommonEngine
from .mysql_engine import MysqlEngine
from .postgres_engine import PostgresEngine
from .sql_server_engine import SqlServerEngine
from .sqlite_engine import SqliteEngine


__all__ = [
    'BasicEngine',
    'CommonEngine',
    'MysqlEngine',
    'PostgresEngine',
    'SqliteEngine',
    'SqlServerEngine'
]



from abc import abstractmethod
from typing import Callable, Tuple


class EngineInterface:
    @abstractmethod
    def make_select(self) -> 'SelectQuery':
        raise NotImplementedError('Must override make_select')

    @abstractmethod
    def make_insert(self) -> 'InsertQuery':
        raise NotImplementedError('Must override make_insert')

    @abstractmethod
    def make_update(self) -> 'UpdateQuery':
        raise NotImplementedError('Must override make_update')
    
    @abstractmethod
    def make_delete(self) -> 'DeleteQuery':
        raise NotImplementedError('Must override make_delete')

    @abstractmethod
    def escape_identifier(self, identifier: str) -> str:
        raise NotImplementedError('Must override escape_identifier')

    @abstractmethod
    def escape_like(self, parameter: str) -> str:
        raise NotImplementedError('Must override escape_like')

    @abstractmethod
    def extract_params(self) -> Callable:
        raise NotImplementedError('Must override extract_params')

    @abstractmethod
    def extract_sql(self) -> Callable:
        raise NotImplementedError('Must override extract_sql')

    @abstractmethod
    def flatten_params(self, *args: 'StatementInterface') -> Tuple:
        raise NotImplementedError('Must override flatten_params')

    @abstractmethod
    def flatten_sql(self, separator: str, *args: 'StatementInterface') -> str:
        raise NotImplementedError('Must override flatten_sql')

    @abstractmethod
    def export_parameter(self, param) -> str:
        raise NotImplementedError('Must override export_parameter')

    @abstractmethod
    def get_parameter_placeholder(self) -> str:
        raise NotImplementedError('Must override get_parameter')



from sql_smith.functions import identify


class HasReturningMixin:
    def returning(self, column: str):
        self._returning = identify(column)
        return self

    def _apply_returning(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('RETURNING {}', self._returning) if self._returning else query



from typing import Dict, Any

from sql_smith.functions import express, identify, listing, identify_all, param_all
from sql_smith.query import AbstractQuery


class InsertQuery(AbstractQuery):
    """Implements the INSERT query."""
    def __init__(self, engine: 'EngineInterface'):
        AbstractQuery.__init__(self, engine)
        self._into = None
        self._columns = None
        self._values = []

    def into(self, table: str) -> 'InsertQuery':
        """Sets the table."""
        self._into = identify(table)
        return self

    def map(self, column_values: Dict[str, Any]):
        """Maps a dictionary to columns and values."""
        return self.columns(*column_values.keys()).values(*column_values.values())

    def columns(self, *columns) -> 'InsertQuery':
        """Sets the columns to insert."""
        self._columns = listing(identify_all(*columns))
        return self

    def values(self, *values) -> 'InsertQuery':
        """Appends values."""
        self._values.append(express('({})', listing(param_all(*values))))
        return self

    def as_expression(self) -> 'ExpressionInterface':
        query = self.start_expression()
        query = self.__apply_into(query)
        query = self.__apply_columns(query)
        query = self.__apply_values(query)

        return query

    def start_expression(self) -> 'ExpressionInterface':
        return express('INSERT')

    def __apply_into(self, query):
        return query.append('INTO {}', self._into) if self._into else query

    def __apply_columns(self, query):
        return query.append('({})', self._columns) if self._columns else query

    def __apply_values(self, query):
        return query.append('VALUES {}', listing(self._values)) if self._values else query



from sql_smith.functions import order, listing


class HasOrderByMixin:
    def order_by(self, column: str = None, direction: str = None):
        if column:
            self._order_by.append(order(column, direction))
            return self

        self._order_by = []
        return self

    def _apply_order_by(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('ORDER BY {}', listing(self._order_by)) if len(self._order_by) > 0 else query



from sql_smith.engine import BasicEngine
from sql_smith.query.sql_server.select_query import SelectQuery
from sql_smith.query.sql_server.delete_query import DeleteQuery


class SqlServerEngine(BasicEngine):
    """A custom engine for SQL Server SQL dialect."""
    def make_select(self) -> 'SelectQuery':
        """Creates a custom SQL Server SELECT query.

        SQL Server handles OFFSET and LIMIT differently.
        """
        return SelectQuery(self)

    def make_delete(self) -> 'DeleteQuery':
        """Creates a custom SQL Server DELETE query.

        SQL Server handles LIMIT differently.
        """
        return DeleteQuery(self)

    def escape_identifier(self, identifier: str) -> str:
        """Escapes identifiers with brackets."""
        return '[{}]'.format(identifier)

    def escape_like(self, parameter: str) -> str:
        """Escape like argument.

        Also escapes character ranges.
        """
        return super().escape_like(parameter).replace('[', r'\[').replace(']', r'\]')



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class Literal(StatementInterface):
    def __init__(self, value):
        self._value = value

    def sql(self, engine: 'EngineInterface') -> str:
        return str(self._value)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return ()




from sql_smith.capability import CanUnionMixin, HasOrderByMixin
from sql_smith.functions import express
from sql_smith.query import AbstractQuery


class UnionQuery(CanUnionMixin, HasOrderByMixin, AbstractQuery):
    """Implements a union query."""
    def __init__(self, engine: 'EngineInterface', left: 'StatementInterface', right: 'StatementInterface'):
        super().__init__(engine)
        self._order_by = []
        self._all = False
        self._left = left
        self._right = right

    def all(self, state: bool = True) -> 'UnionQuery':
        """Sets ALL."""
        self._all = state
        return self

    def as_expression(self) -> 'ExpressionInterface':
        query = self.start_expression()
        query = self.__apply_all(query)
        query = self.__apply_right(query)
        query = self._apply_order_by(query)
        return query

    def start_expression(self) -> 'ExpressionInterface':
        return express('{} UNION', self._left)

    def __apply_all(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('ALL') if self._all else query

    def __apply_right(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('{}', self._right)



from .criteria import Criteria
from .expression import Expression
from .identifier import Identifier
from .like_begins import LikeBegins
from .like_contains import LikeContains
from .like_ends import LikeEnds
from .qualified_identifier import QualifiedIdentifier
from .listing import Listing
from .literal import Literal

__all__ = [
    'Criteria',
    'Expression',
    'Identifier',
    'LikeBegins',
    'LikeContains',
    'LikeEnds',
    'Listing',
    'Literal',
    'QualifiedIdentifier'
]



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class NullParameter(StatementInterface):
    def sql(self, engine: 'EngineInterface') -> str:
        return engine.export_parameter(None)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return ()



from sql_smith.functions import criteria
from sql_smith.partial import LikeBegins, LikeContains, LikeEnds


class LikeBuilder:
    def __init__(self, statement: 'StatementInterface'):
        self._statement = statement

    def begins(self, value: str) -> 'CriteriaInterface':
        return self.__like(LikeBegins(value))

    def not_begins(self, value: str) -> 'CriteriaInterface':
        return self.__not_like(LikeBegins(value))

    def contains(self, value: str) -> 'CriteriaInterface':
        return self.__like(LikeContains(value))

    def not_contains(self, value: str) -> 'CriteriaInterface':
        return self.__not_like(LikeContains(value))

    def ends(self, value: str) -> 'CriteriaInterface':
        return self.__like(LikeEnds(value))

    def not_ends(self, value: str) -> 'CriteriaInterface':
        return self.__not_like(LikeEnds(value))

    def __like(self, value: 'StatementInterface') -> 'CriteriaInterface':
        return criteria('{} LIKE {}', self._statement, value)

    def __not_like(self, value: 'StatementInterface') -> 'CriteriaInterface':
        return criteria('{} NOT LIKE {}', self._statement, value)



from typing import Tuple


class Query:
    """The result of a compile."""
    def __init__(self, sql: str, params: Tuple):
        self._sql = sql
        self._params = params

    @property
    def sql(self) -> str:
        """Returns the compiled SQL statement as a string."""
        return self._sql

    @property
    def params(self) -> Tuple:
        """Returns all parameters of the SQL statement."""
        return self._params



from typing import Tuple, Union, List

from sql_smith.interfaces import ExpressionInterface, StatementInterface
from sql_smith.partial import Expression, QualifiedIdentifier, Identifier, Literal, Listing, Criteria
from sql_smith.partial.parameter import Parameter


def __is_statement(value) -> bool:
    return isinstance(value, StatementInterface)


def param(value) -> 'StatementInterface':
    """Create a parameter.

    >>> func('POINT', param(1), param(2))  # POINT(? , ?)
    """
    if __is_statement(value):
        return value
    return Parameter.create(value)


def param_all(*values):
    return tuple(map(param, values))


def express(pattern: str, *args):
    """Create an expression.

    >>> express('{} + 1', identify('visit'))  # "visit" + 1
    """
    return Expression(pattern, *param_all(*args))


def alias(field_name, field_alias: str) -> 'ExpressionInterface':
    """Create an alias for a column or a function.

    >>> alias('users.id', 'uid')  # "users"."id" AS "uid"
    """
    return express('{} AS {}', identify(field_name), identify(field_alias))


def listing(values: Union[Tuple, List], separator: str = ', ') -> Listing:
    """Create a listing.

    >>> listing((1, 1, 2, 3, 5))  # ?, ?, ?, ?, ?
    >>> listing(identify_all('id', 'username', 'email'))  # "id", "username", "email"
    """
    return Listing(separator, *param_all(*values))


def func(function: str, *args) -> 'ExpressionInterface':
    """Create a function.

    >>> func('COUNT', 'user.id')  # COUNT("users"."id")
    """
    return express('{}({{}})'.format(function), listing(identify_all(*args)))


def literal(value) -> 'StatementInterface':
    """Create a literal.
    """
    if __is_statement(value):
        return value
    return Literal(value)


def criteria(pattern: str, *args) -> 'CriteriaInterface':
    """Create a criteria.

    >>> c = criteria(
    >>>     "{} = {}",
    >>>     func(
    >>>         'YEAR',
    >>>         identify('start_date')
    >>>     ),
    >>>     literal(2021)
    >>> )  # YEAR("start_date") = 2021
    """
    return Criteria(express(pattern, *args))


def on(left: str, right: str):
    """Create an on clause.
    """
    return criteria('{} = {}', identify(left), identify(right))


def order(column, direction: str = None) -> 'StatementInterface':
    """Create an order clause."""
    if direction is None:
        return identify(column)
    return express('{{}} {}'.format(direction.upper()), identify(column))


def group(c: 'CriteriaInterface') -> 'CriteriaInterface':
    """Create a group of criteria.

    >>> group(
    >>>     field('username').eq('tom')
    >>>     .or_(field('first_name').eq('Tom'))
    >>> ).and_(
    >>>     field('is_active').eq(1)
    >>> )
    >>> # ("username" = ? OR "first_name" = ?) AND "is_active" = ?
    """
    return criteria('({})', c)


def field(name):
    """Starts a criteria for a column. Use it to create a condition.

    >>> field('users.id').eq(100)  # "users".id = ?
    """
    from sql_smith.builder import CriteriaBuilder
    return CriteriaBuilder(identify(name))


def search(name):
    """Start a LIKE clause.

    >>> search('username').contains('admin')  # "username" LIKE '%admin%'
    """
    from sql_smith.builder import LikeBuilder
    return LikeBuilder(identify(name))


def identify_all(*names) -> Tuple:
    """Identify all names.

    >>> identify_all('id', 'username')  # ("id", "username")
    """
    return tuple(map(identify, names))


def identify(name) -> 'StatementInterface':
    """Identify a name.

    >>> identify('users.id')  # "users"."id"
    """
    if __is_statement(name):
        return name

    if name.find('.') != -1:
        return QualifiedIdentifier(*identify_all(*name.split('.')))

    if name == '*':
        return Literal(name)

    return Identifier(name)



from .abstract_query import AbstractQuery
from .delete_query import DeleteQuery
from .insert_query import InsertQuery
from .query import Query
from .select_query import SelectQuery
from .update_query import UpdateQuery
from .union_query import UnionQuery

__all__ = [
    'AbstractQuery',
    'DeleteQuery',
    'InsertQuery',
    'Query',
    'SelectQuery',
    'UpdateQuery',
    'UnionQuery'
]



from sql_smith.engine import BasicEngine


class SqliteEngine(BasicEngine):
    """A custom engine for Sqlite SQL dialect."""
    def export_parameter(self, param) -> str:
        """Export a parameter.

        SQLite doesn't support boolean storage class.
        True is translated to 1, False to 0.
        """
        if type(param) == bool:
            return str(int(param))
        return super().export_parameter(param)



from sql_smith.query import SelectQuery as BaseSelectQuery


class SelectQuery(BaseSelectQuery):
    """Mysql SelectQuery that supports SQL_CALC_FOUND_ROWS."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__(engine)
        self._calc_found_rows = False

    def calc_found_rows(self, status: bool) -> 'MySqlSelectQuery':
        self._calc_found_rows = status
        return self

    def start_expression(self) -> 'ExpressionInterface':
        query = super().start_expression()
        if self._calc_found_rows:
            query = query.append('SQL_CALC_FOUND_ROWS')
        return query



from functools import reduce
from typing import Callable, Tuple

from sql_smith.interfaces import EngineInterface, StatementInterface
from sql_smith.query import SelectQuery, UpdateQuery, DeleteQuery, InsertQuery


class BasicEngine(EngineInterface):
    """Basic engine

    A default implementation of an engine.
    """
    def make_select(self) -> 'SelectQuery':
        """Creates a SelectQuery."""
        return SelectQuery(self)

    def make_insert(self) -> 'InsertQuery':
        """Creates an InsertQuery."""
        return InsertQuery(self)

    def make_update(self) -> 'UpdateQuery':
        """Creates an UpdateQuery."""
        return UpdateQuery(self)
    
    def make_delete(self) -> 'DeleteQuery':
        """Creates a DeleteQuery."""
        return DeleteQuery(self)

    def escape_identifier(self, identifier: str) -> str:
        """Escapes an identifier.

        The default implementation returns the identifier as is.
        """
        return identifier

    def escape_like(self, parameter: str) -> str:
        """Escapes the LIKE argument

        A backslash is used to escape wildcards.
        Standard wildcards are underscore and percent sign.
        """
        return parameter.replace('\\', '\\\\').replace('%', '\\%').replace('_', '\\_')

    def export_parameter(self, param) -> str:
        """Export a parameter.

        A boolean will be exported as true or false. A None value will be exported as NULL.
        """
        if isinstance(param, bool):
            return 'true' if param else 'false'
        if param is None:
            return 'NULL'
        return str(param)

    def extract_params(self) -> Callable[['StatementInterface'], Tuple]:
        """Create a lambda to extract parameters."""
        return lambda statement: statement.params(self)

    def extract_sql(self) -> Callable[['StatementInterface'], str]:
        """Create a lambda to extract SQL."""
        return lambda statement: statement.sql(self)

    def flatten_params(self, *args: 'StatementInterface') -> Tuple:
        """Create a tuple with all parameters found."""
        # map with extract_params returns a list with tuples,
        # the reduce method is used to merge all these tuples into one big tuple
        return reduce(lambda x, y: (x + y), map(self.extract_params(), args), ())

    def flatten_sql(self, separator: str, *args: 'StatementInterface') -> str:
        """Transform StatementInterface arguments to a string."""
        return separator.join(list(map(self.extract_sql(), args)))

    def get_parameter_placeholder(self):
        return '?'



from sql_smith.query import InsertQuery as BaseInsertQuery


class InsertQuery(BaseInsertQuery):
    """MySql InsertQuery that supports IGNORE."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__(engine)
        self._ignore = False

    def ignore(self, status: bool) -> 'MySqlSelectQuery':
        self._ignore = status
        return self

    def start_expression(self) -> 'ExpressionInterface':
        query = super().start_expression()
        if self._ignore:
            query = query.append('IGNORE')
        return query



from sql_smith.functions import literal
from sql_smith.query import SelectQuery as BaseSelectQuery


class SelectQuery(BaseSelectQuery):
    """Sql Server SelectQuery to support offset and limit."""
    def _apply_offset(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        if self._offset is None or self._limit is None:
            return query
        return query.append('OFFSET {} ROWS FETCH NEXT {} ROWS ONLY', literal(self._offset), literal(self._limit))

    def _apply_limit(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query



from typing import Union, Dict, Any

from sql_smith.engine import BasicEngine


class QueryFactory:
    """Factory class for creating queries
    """
    def __init__(self, engine: 'EngineInterface' = None):
        if engine is None:
            self._engine = BasicEngine()
        else:
            self._engine = engine

    def select(self, *columns: Union[str, 'StatementInterface']) -> 'SelectQuery':
        """Create a SELECT query
        """
        query = self._engine.make_select()
        if len(columns) == 0:
            return query

        return query.columns(*columns)

    def select_distinct(self, *columns: Union[str, 'StatementInterface']) -> 'SelectQuery':
        """Create a SELECT DISTINCT query
        """
        return self.select(*columns).distinct()

    def insert(self, table: str, column_values: Dict[str, Any] = None) -> 'InsertQuery':
        """Create an INSERT query
        """
        query = self._engine.make_insert().into(table)
        if column_values:
            query = query.map(column_values)

        return query

    def delete(self, table: str) -> 'DeleteQuery':
        """Create a DELETE query
        """
        return self._engine.make_delete().from_(table)

    def update(self, table: str, value_set=None) -> 'UpdateQuery':
        """Create an UPDATE query
        """
        query = self._engine.make_update().table(table)
        if value_set:
            query.set(value_set)

        return query



from typing import Union

from sql_smith.functions import identify_all, listing


class HasFromMixin:
    def from_(self, *tables: Union[str, 'ExpressionInterface']):
        self._from = identify_all(*tables)
        return self

    def add_from(self, *tables: Union[str, 'ExpressionInterface']):
        self._from = (*self._from, *identify_all(*tables))
        return self

    def _apply_from(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        if self._from:
            return query.append('FROM {}', listing(self._from))
        return query



from .criteria_interface import CriteriaInterface
from .engine_interface import EngineInterface
from .expression_interface import ExpressionInterface
from .query_interface import QueryInterface
from .statement_interface import StatementInterface

__all__ = [
    'CriteriaInterface',
    'EngineInterface',
    'ExpressionInterface',
    'QueryInterface',
    'StatementInterface'
]



from sql_smith.engine import BasicEngine


class CommonEngine(BasicEngine):
    """An engine that supports common SQL standard."""
    def escape_identifier(self, identifier: str) -> str:
        """Escapes identifiers by putting quotes around it."""
        return '"{}"'.format(identifier)



from abc import abstractmethod
from typing import Tuple


class StatementInterface:
    @abstractmethod
    def sql(self, engine: 'EngineInterface') -> str:
        raise NotImplementedError('sql must be implemented')

    @abstractmethod
    def params(self, engine: 'EngineInterface') -> Tuple:
        raise NotImplementedError('params must be implemented')



#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'sql-smith',
        version = '1.0.1',
        description = 'sql-smith is an SQL query builder with zero dependencies and a fluent interface.',
        long_description = '=========\nsql-smith\n=========\n\n**sql-smith** is an SQL query builder with zero dependencies and a fluent interface.\n\n    The sentence above is, beside the name, a copy from the website of the PHP library\n    Latitude_, for the simple reason that this Python module is a port of Latitude.\n\nRead the full `documentation <https://fbraem.github.io/sql-smith>`_.\n\nInstallation\n************\n\n.. code-block:: sh\n\n    $ pip install sql-smith\n\nQuick Start\n***********\n\nQueryFactory is a factory to create a **SELECT**, **INSERT**, **UPDATE** or **DELETE** query.\nUse the fluent interface of the queries to complete the query.\n\n.. code-block:: python\n\n    from sql_smith import QueryFactory\n    from sql_smith.engine import CommonEngine\n    from sql_smith.functions import field\n    \n    factory = QueryFactory(CommonEngine())\n    query = factory \\\n        .select(\'id\', \'username\') \\\n        .from_(\'users\') \\\n        .where(field(\'id\').eq(5)) \\\n        .compile()\n    \n    print(query.sql)  # SELECT "id", "username" FROM "users" WHERE "id" = ?\n    print(query.params)  # (5)\n\nWhen the query is ready, compile it. The return value of compile is a Query class instance\nwith two properties: sql and params. Use these properties to pass the query to a database.\n\n.. code-block:: python\n\n    import sqlite3\n    \n    db = sqlite3.connect(\'test.db\')\n    cur = db.cursor()\n\n    for row in cur.execute(query.sql, query.params):\n        print(row)\n\n.. _Latitude: https://latitude.shadowhand.com/\n',
        long_description_content_type = 'text/x-rst',
        classifiers = [
            'License :: OSI Approved :: MIT License',
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Operating System :: OS Independent'
        ],
        keywords = 'sql-smith query builder querybuilder sql mysql sqlite postgres sqlserver database',

        author = 'Franky Braem',
        author_email = 'franky.braem@gmail.com',
        maintainer = '',
        maintainer_email = '',

        license = 'MIT',

        url = 'https://github.com/fbraem/sql-smith',
        project_urls = {},

        scripts = [],
        packages = [
            'sql_smith',
            'sql_smith.builder',
            'sql_smith.capability',
            'sql_smith.engine',
            'sql_smith.interfaces',
            'sql_smith.partial',
            'sql_smith.partial.parameter',
            'sql_smith.query',
            'sql_smith.query.mysql',
            'sql_smith.query.postgres',
            'sql_smith.query.sql_server'
        ],
        namespace_packages = [],
        py_modules = [],
        entry_points = {},
        data_files = [],
        package_data = {},
        install_requires = [],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '',
        obsoletes = [],
    )



from sql_smith.functions import literal
from sql_smith.query import DeleteQuery as BaseDeleteQuery


class DeleteQuery(BaseDeleteQuery):
    """Sql Server DeleteQuery to support LIMIT."""
    def start_expression(self) -> 'ExpressionInterface':
        query = super().start_expression()
        if self._limit is None:
            return query
        return query.append('TOP({})', literal(self._limit))

    def _apply_limit(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query



from sql_smith.functions import criteria, listing


class CriteriaBuilder:
    def __init__(self, statement: 'StatementInterface'):
        self._statement = statement

    def between(self, start, end) -> 'CriteriaInterface':
        return criteria('{} BETWEEN {} AND {}', self._statement, start, end)

    def not_between(self, start, end) -> 'CriteriaInterface':
        return criteria('{} NOT BETWEEN {} AND {}', self._statement, start, end)

    def in_(self, *args) -> 'CriteriaInterface':
        return criteria('{} IN ({})', self._statement, listing(args))

    def not_in(self, *args) -> 'CriteriaInterface':
        return criteria('{} NOT IN ({})', self._statement, listing(args))

    def eq(self, value) -> 'CriteriaInterface':
        return criteria('{} = {}', self._statement, value)

    def not_eq(self, value) -> 'CriteriaInterface':
        return criteria('{} != {}', self._statement, value)

    def gt(self, value) -> 'CriteriaInterface':
        return criteria('{} > {}', self._statement, value)

    def gte(self, value) -> 'CriteriaInterface':
        return criteria('{} >= {}', self._statement, value)

    def lt(self, value) -> 'CriteriaInterface':
        return criteria('{} < {}', self._statement, value)

    def lte(self, value) -> 'CriteriaInterface':
        return criteria('{} <= {}', self._statement, value)

    def is_null(self) -> 'CriteriaInterface':
        return criteria('{} IS NULL', self._statement)

    def is_not_null(self) -> 'CriteriaInterface':
        return criteria('{} IS NOT NULL', self._statement)



from typing import Dict, Any, Union

from sql_smith.capability import HasWhereMixin
from sql_smith.functions import identify, param, express, listing
from sql_smith.query import AbstractQuery


class UpdateQuery(HasWhereMixin, AbstractQuery):
    """Implements an UPDATE query."""
    def __init__(self, engine: 'EngineInterface'):
        super().__init__(engine)
        self._table = None
        self._set = None
        self._where = []

    def table(self, table: Union[str, 'StatementInterface']) -> 'UpdateQuery':
        """Sets the table."""
        self._table = identify(table)
        return self

    def set(self, value_dict: Dict[str, Any]):
        """Sets the column and values with a dictionary."""
        self._set = listing(list(map(
            lambda k, v: express('{} = {}', identify(k), param(v)),
            value_dict.keys(),
            value_dict.values()
        )))
        return self

    def as_expression(self) -> 'ExpressionInterface':
        query = self.start_expression()
        query = self.__apply_table(query)
        query = self.__apply_set(query)
        query = self._apply_where(query)

        return query

    def start_expression(self) -> 'ExpressionInterface':
        return express('UPDATE')

    def __apply_table(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('{}', self._table) if self._table else query

    def __apply_set(self, query: 'ExpressionInterface') -> 'ExpressionInterface':
        return query.append('SET {}', self._set) if self._set else query



from typing import Tuple

from sql_smith.interfaces import CriteriaInterface


class Criteria(CriteriaInterface):
    def __init__(self, expression: 'ExpressionInterface'):
        self._expression = expression

    def and_(self, right: 'CriteriaInterface') -> 'CriteriaInterface':
        return Criteria(self._expression.append('AND {}', right))

    def or_(self, right: 'CriteriaInterface') -> 'CriteriaInterface':
        return Criteria(self._expression.append('OR {}', right))

    def sql(self, engine: 'EngineInterface') -> str:
        return self._expression.sql(engine)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return self._expression.params(engine)



from typing import Tuple

from sql_smith.interfaces import StatementInterface


class Listing(StatementInterface):
    def __init__(self, separator: str, *statements: 'StatementInterface'):
        self._separator = separator
        self._statements = statements

    def sql(self, engine: 'EngineInterface') -> str:
        return engine.flatten_sql(self._separator, *self._statements)

    def params(self, engine: 'EngineInterface') -> Tuple:
        return engine.flatten_params(*self._statements)



from sql_smith.engine import BasicEngine
from sql_smith.query.mysql import SelectQuery, InsertQuery


class MysqlEngine(BasicEngine):
    """A custom engine for MySQL SQL dialect."""
    def make_select(self) -> 'SelectQuery':
        """Creates a custom MySql SELECT query.

        The MySQL select query supports SQL_CALC_FOUND_ROWS.
        """
        return SelectQuery(self)

    def make_insert(self) -> 'InsertQuery':
        """Creates a custom MySql INSERT query.

        The MySQL insert query supports IGNORE.
        """
        return InsertQuery(self)

    def escape_identifier(self, identifier: str) -> str:
        """Escapes the identifier by surrounding it with backticks."""
        return '`{}`'.format(identifier)

    def get_parameter_placeholder(self):
        return '%s'
